/*******************************************************************************
* Implementation of the compiler generator library containing actions used by
* the scanner and parser generated with SLADE
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#include "cgl.h"

#if defined(unix) || defined(WIN32)

#ifndef Debugger
#define MAXDRIVE	1
#define MAXDIR		255
#define MAXFILE		80
#define MAXEXT		40
#define MAXPATH		(MAXDIR+MAXFILE+MAXEXT)
#endif

#ifdef WIN32
int					optind;
char				*optarg;
#else
extern char			*optarg;
extern int			optind;
#endif /* WIN32 */
#endif /* unix || WIN32 */

string				i_lexeme;		/* current token representation */
char				*i_lexeme_ptr;	/* end of current token representation */
char				*i_lexeme_end;	/* end of current token representation */
int					i_ch;			/* current input character*/
integer				i_linenr, i_nr_errors;
char				i_input_name[MAXFILE];

static boolean		AtNewLine, CopyLastLine, LstReset,
					TestScanOpt, TestParsOpt;
static char			InputPath[MAXPATH];
static int			LastLineLen;
static FILE			*InpFp;
static FILE			*LstFp;

static char *Reports[]= {
	/* error reports */ "",										/* i_ERROR    */
	"Access '%s' not allowed here",								/* i_EACCNOTA */
	"Access conflict: %s <-> %s",								/* i_EACCONF  */
	"Cannot change address of %s",								/* i_EADDRESS */
	"Boundary size conflict: %s <-> %s",						/* i_EBNDCONF */
	"Equal boundary sizes %d, not allowed here",				/* i_EBNDNOTA */
	"Double defined identifier: '%s'",							/* i_EDBLDEF  */
	"Dimension conflict: %d <-> %d",							/* i_EDIMCONF */
	"Dimension %d not allowed here",							/* i_EDIMNOTA */
	"Divide by zero",											/* i_EDIVBY0  */
	"VIM code not initialised or already finalised",			/* i_EFINVIM  */
	"File '%s' does not exist or is unreadable",				/* i_EFREAD   */
	"File '%s' not writable, cannot continue",					/* i_EFWRITE  */
	"Illegal argument of emit_jump(): '%s'",					/* i_EILLARG  */
	"Illegal reference argument, %s is read-only",				/* i_EILLREF  */
	"VIM code already initialised",								/* i_EINIVIM  */
	"Scanner input buffer overflow",							/* i_EINPUTBUF*/
	"Kind conflict: %s <-> %s",									/* i_EKINDCONF*/
	"Kind '%s' not allowed here",								/* i_EKINDNOTA*/
	"Length table full (more then %d items)",					/* i_ELENTAB  */
	"Modulo by zero",											/* i_EMODBY0  */
	"%s is read-only, cannot be accessed by address",			/* i_ENOADDR  */
	"No field '%s' in record '%s'",								/* i_ENOFIELD */
	"Insertion of identifier '%s' failed, no current scope",	/* i_ENOIDINS */
	"%s cannot be accessed indirectly",							/* i_ENOIND   */
	"Scanner could not find keyword: '%s'",						/* i_ENOKEYW  */
	"No multi-dimensional version of operation '%s' available",	/* i_ENOMDOP  */
	"Out of memory at line %d in %s, cannot continue",          /* i_ENOMEM   */
	"No current scope, exit_scope() failed",					/* i_ENOSCOPE */
	"Scanner has no initial state for input character '%c'",	/* i_ENOSTATE */
	"Cannot store, %s is read-only",							/* i_ENOSTORE */
	"Identifier '%s' not found",								/* i_ENOTFOUND*/
	"Scanner skipped unknown token, lexeme = '%s'",				/* i_ENOTOKEN */
	"Cannot emit operation '%s' without arguments",				/* i_EOPARGS  */
	"Lower bound %d exceeds upper bound %d",					/* i_ERANGE   */
	"No VIM read instruction for type '%s'",					/* i_ERDTYPE  */
	"String table full (more then %d items)",					/* i_ESTRTAB  */
	"Parser expected (non)terminal '%s'",						/* i_ETERMEXP */
	"Too few %s",												/* i_ETOOFEW  */
	"Too many %s",												/* i_ETOOMANY */
	"Type conflict: %s <-> %s",									/* i_ETYPECONF*/
	"Type '%s' not allowed here",								/* i_ETYPENOTA*/
	"No VIM write instruction for type '%s'",					/* i_EWRTYPE  */

	/* warning reports */ "",									/* i_WARNING  */
	"Possible dangling pointer reference",						/* i_WDANGLE  */
	"Parser inserted unnamed syntactical unit (group)",			/* i_WGRPINS  */
	"Procedure '%s' defined but never called",					/* i_WNOCALL  */
	"Variable '%s' might be used before set",					/* i_WNOTSET  */
	"Identifier '%s' defined but never set or used",			/* i_WNOTUSED */
	"Parser deleted terminal '%s' (lexeme = '%s')",				/* i_WTERMDEL */
	"Parser inserted (non)terminal '%s'",						/* i_WTERMINS */

	/* trace reports */ "",										/* i_TRACE    */
	"- input file     : %s",									/* i_TINP	*/
	"- listing file   : %s",									/* i_TLST     */
	"- errors         : %d error(s) detected",					/* i_TNERRS   */
	"%18d VIM instruction(s)",									/* i_TNITAB   */
	"%18d length table item(s)",								/* i_TNLTAB   */
	"%18d string table item(s)",								/* i_TNSTAB   */
	"",															/* i_TPARSTST */
	"",															/* i_TSCANTST */
	"Usage: %s [-ts] [-tp] [-h] file\n"
	"-ts:   test scanner\n"
	"-tp:   test parser\n"
	"-h:    print this information",							/* i_TUSAGE   */
	"Parser %s generated by SLADE v2.9",						/* i_TVERSION */
	"- vim code       : %s",									/* i_TVIM	*/
};

void
i_report(i_ReportNumber rnr, ...)
{
	va_list	args;
	char	buf[2*i_STRINGLEN+1];
	char	*report, *bufp;

	va_start(args, rnr);
	if ((rnr == i_TPARSTST && !TestParsOpt)
	|| (rnr == i_TSCANTST && !TestScanOpt))
		return;


	if (rnr != i_TRACE && rnr != i_TPARSTST && rnr != i_TSCANTST) {
		if (rnr == i_ERROR || rnr == i_WARNING)
			report= va_arg(args, char *);
		else
			report= Reports[rnr];
		buf[0]= 0;
		if (rnr < i_WARNING) {								/* error report */
			strcpy(buf, "Error: ");
			i_nr_errors++;
		} else if (rnr < i_TRACE)							/* warning report */
			strcpy(buf, "Warning: ");

		bufp= buf + strlen(buf);;
	} else {
		report= va_arg(args, char *);
		bufp= buf;
	}
	vsprintf(bufp, report, args);
	va_end(args);

	if (rnr == i_TNERRS && i_nr_errors > 0)
		strcat(buf, ", cannot continue.");

	if (LstFp != 0) {
		if (!LstReset) {
			fseek(LstFp, (long)(i_lexeme_end - i_lexeme_ptr), SEEK_CUR);
			LstReset= true;
		}
		if (rnr < i_TRACE) {
			fprintf(LstFp, "\n*** %s", buf);
			CopyLastLine= true;
		} else if ((rnr == i_TPARSTST && TestParsOpt)
		|| (rnr == i_TSCANTST && TestScanOpt)) {
			fprintf(LstFp, "\n%s", buf);
			CopyLastLine= true;
			return;
		}
	}

#ifdef Debugger
	if (rnr >= i_TRACE) {
		char	*s, *bufp;

		bufp= buf;
		if ((s= strchr(bufp, '\n')) != 0) {
			do {
				*s= 0;
				i_add_list_item(IdOutput, bufp);
				bufp= s+1;
			} while ((s= strchr(bufp, '\n')) != 0);
		}
		if (*bufp)
			i_add_list_item(IdOutput, bufp);
	} else
		i_add_list_item(IdOutput, buf);
#else
	if (i_linenr > 0 && rnr < i_TRACE)
		fprintf(stderr, "%s:%d: ", InputPath, i_linenr);

	if (rnr != i_TRACE) {
		strcat(buf, "\n");
		fputs(buf, stderr);
	} else 
		fputs(buf, stdout);

	if (rnr == i_ENOMEM)
		exit(1);
#endif /* Debugger */
}

/* Default actions for the standard types */

void
i_integer(integer *integer_value)
{
	*integer_value= 0;
}

void
i_real(real *real_value)
{
	*real_value= 0.0;
}

void
i_boolean(boolean *boolean_value)
{
	*boolean_value= false;
}

void
i_char(char *char_value)
{
	*char_value= 0;
}

void
i_string(string *string_value)
{
	(*string_value)[0]= 0;
}

char *
i_read_line(char *s, int len, char *prompt)
{
#ifdef Debugger
	return i_read_line_dialog(s, len, prompt);
#else
	if (prompt && *prompt)
		fputs(prompt, stdout);
	if (fgets(s, len, stdin))
		s[strlen(s)-1]= 0;	/* remove trailing newline */
	return s;
#endif /* Debugger */
}

#ifdef Debugger
void
i_reset_cgl_files(void)
{
	if (LstFp != 0) {
		fclose(LstFp);
		LstFp= 0;
	}

	if (InpFp != 0) {
		fclose(InpFp);
		InpFp= 0;
	}
}
#endif /* Debugger */

/* Standard scanner actions */

char *
get_repr(string repr)
{
	return repr;
}

char *
convert_to_upper(string repr)
{
	static string	result;
	char			*s;

	strcpy(result, repr);
	for (s= result; *s; s++) 
		if (islower(*s))
			*s= toupper(*s);

	return result;
}

char *
convert_to_lower(string repr)
{
	static string	result;
	char			*s;

	strcpy(result, repr);
	for (s= result; *s; s++) 
		if (isupper(*s))
			*s= tolower(*s);

	return result;
}

char *
delete_spaces(string repr)
{
	static string	result;
	char			*s= result;
	int				i, start, leng;

	*s= 0;
	leng= strlen(repr);

	i= 0;
	while (i < leng) {
		start= i;
		while (i < leng && repr[i] != ' ')
			i++;

		strncat(s, repr+start, i-start);

		while (i < leng && repr[i] == ' ')
			i++;
	}
	return result;
}

integer int_encode(string repr)
{
	return atoi(repr);
}

char char_encode(string repr)
{
	if (strlen(repr) >= 3) {
		if (repr[1] == '\\') {
			switch (repr[2]) {
			case 'a': return '\a';
			case 'b': return '\b';
			case 'f': return '\f';
			case 'n': return '\n';
			case 'r': return '\r';
			case 't': return '\t';
			case 'v': return '\v';
			default: return repr[2];
			}
		} else
			return repr[1];
	}
	return repr[0];
}

boolean
bool_encode(string repr)
{
	if (repr[0] == 'T' || repr[0] == 't')
		return true;

	return false;
}

real
real_encode(string repr)
{
	return atof(repr);
}

char *
string_encode(string repr)
{
	static string	result;

	/* remove first and last character (e.g. quotes) */
	strcpy(result, repr+1);
	result[strlen(result) - 1]= 0;
	return result;
}

void
i_init_lexeme(void)
{
	if (i_ch == EOF)
		return;

	if (i_lexeme_end) {
		if (!LstReset && LstFp) {
			fseek(LstFp, (long)(i_lexeme_end - i_lexeme_ptr), SEEK_CUR);
			LstReset= true;
		}
		i_reset_input();
	}
	i_lexeme_ptr= i_lexeme;
	i_lexeme_end= 0;
}

void
i_reset_input(void)
{
	if (InpFp == 0)
		return;
	
	while (i_lexeme_ptr != i_lexeme_end) {
		ungetc(*(--i_lexeme_ptr), InpFp);
		if (*i_lexeme_ptr == '\n')
			AtNewLine= false;
		else
			LastLineLen--;
	}
}

void
i_next_character(void)
{
	static string	lastline;

	if (i_ch == EOF || InpFp == 0 || LstFp == 0)
		return;

	if ((i_ch= fgetc(InpFp)) != EOF) {
		*i_lexeme_ptr++= i_ch;
		if (CopyLastLine)
			fputc('\n', LstFp);

		if (AtNewLine) {
			fprintf(LstFp, "%3d ", ++i_linenr);
			sprintf(lastline, "%3d ", i_linenr);
			LastLineLen= strlen(lastline);
			AtNewLine= false;
		} else if (CopyLastLine) {
			lastline[LastLineLen]= 0;
			fprintf(LstFp, "\n%s", lastline);
		}

		CopyLastLine= false;
		fputc(i_ch, LstFp);
		LstReset= false;
		if (i_ch != '\n') {
			if (LastLineLen < i_STRINGLEN)
				lastline[LastLineLen++]= i_ch;
		} else
			AtNewLine= true;
	} else
		*i_lexeme_ptr++= 0;

	if (i_lexeme_ptr - i_lexeme >= i_STRINGLEN) {
		i_report(i_EINPUTBUF);
		*i_lexeme_ptr= 0;
		i_lexeme_ptr= i_lexeme;
	}
}

int
i_finish(void)
{
	if (LstFp != 0) {
		fputc('\n', LstFp);
		fclose(LstFp);
		LstFp= 0;
	}
	i_report(i_TNERRS, i_nr_errors);
	return i_nr_errors;
}

#if defined(unix) || defined(WIN32)
static void
fnsplit(char *path, char *drive, char *dir, char *fname, char *ext)
{
	char	*sep, *s;

	if (drive)
		*drive= 0;

#if defined(unix)
	if ((sep= strrchr(path, '/')) != 0) {
#else /* WIN32 */
	if ((sep= strrchr(path, '\\')) != 0) {
#endif
		if (dir) {
			int		len= sep - path + 1;
			strncpy(dir, path, len);
			dir[len]= 0;
		}

		if (fname)
			strcpy(fname, sep+1);
	} else {
		if (dir)
			*dir= 0;
		if (fname)
			strcpy(fname, path);
	}

	if (!fname)
		s= path;
	else
		s= fname;

	if ((sep= strrchr(s, '.')) != 0) {
		if (fname)
			*sep= 0;

		if (ext)
			strcpy(ext, sep+1);
	} else if (ext)
		*ext= 0;
}

#endif /* unix || WIN32 */

#ifdef unix
#include <unistd.h>
#else /* msdos or mswindows */

#include <dos.h>

int				optind= 1;		/* index of which argument is next	*/
char			*optarg;		/* pointer to argument of current option */
 
static char		*letP= 0;	/* remember next option char's location */
static char		SW= 0;			/* DOS switch character, either '-' or '/' */

int
getopt(int argc, char *argv[], char *options)
{
	unsigned char ch;
	char *optP;
  
#ifdef WIN32
	SW= '-';
#else /* WIN16 or DOS */
	if (SW == 0) {
		/* get SW using dos call 0x37 */
		_AX= 0x3700;
		geninterrupt(0x21);
		SW= _DL;
	}
#endif
 
	if (argc > optind) {
		if (letP == 0) {
			if ((letP= argv[optind]) == 0 || (*letP != SW && *letP != '-'))
				goto gopEOF;
			letP++;
			if (*letP == SW || *letP == '-') {
				optind++;
				goto gopEOF;
			}
		}
		if (0 == (ch= *(letP++))) {
			optind++;  goto gopEOF;
		}
		if (':' == ch  ||  (optP= strchr(options, ch)) == 0)  
			goto gopError;
		if (':' == *(++optP)) {
			optind++;
			if (0 == *letP) {
				if (argc <= optind)  goto  gopError;
				letP= argv[optind++];
			}
			optarg= letP;
			letP= 0;
		} else {
			if (0 == *letP) {
				optind++;
				letP= 0;
			}
			optarg= 0;
		}
		return ch;
	}
gopEOF:
	optarg= letP= 0;  
	return EOF;
 
gopError:
	optarg= 0;
	return ('?');
}
#endif

static boolean
read_parameters(int ac, char **av)
{
	int		c;

	optind= 1;
	TestScanOpt= false;
	TestParsOpt= false;
	while ((c= getopt(ac, av, "t:h")) != -1)
		switch (c) {
		case 't':
			if (*optarg == 'p')
				TestParsOpt= true;
			else if (*optarg == 's')
				TestScanOpt= true;
			else
				i_report(i_TUSAGE, av[0]);
			break;

		case 'h':
		case '?':
			i_report(i_TUSAGE, av[0]);
			return false;
		}
	
	if (optind < ac) {
		char	dir[MAXDIR], ext[MAXEXT];

		if ((InpFp= fopen(av[optind], "rt")) == 0) {
			fnsplit(av[optind], 0, dir, i_input_name, ext);
			InputPath[0]= 0;
			if (dir[0])
				strcpy(InputPath, dir);
			strcat(InputPath, i_input_name);
			if ((InpFp= fopen(InputPath, "rt")) == 0) {
				i_report(i_EFREAD, InputPath);
				return false;
			}
		} else {
			fnsplit(av[optind], 0, 0, i_input_name, 0);
			strcpy(InputPath, av[optind]);
		}
		optind++;
	} else {
		i_report(i_TUSAGE, av[0]);
		return false;
	}
	i_report(i_TINP, InputPath);
	return true;
}

boolean
i_init(int ac, char **av)
{
	char fname[MAXFILE+MAXEXT+2];

	i_ch= 0;
	i_lexeme_ptr= 0;
	i_lexeme_end= 0;
	i_linenr= 0;
	i_nr_errors= 0;
	LstReset= false;
	CopyLastLine= false;
	AtNewLine= true;
	i_report(i_TVERSION, av[0]);

	if (!read_parameters(ac, av))
		return false;
	
	sprintf(fname, "%s.lst", i_input_name);
	if ((LstFp= fopen(fname, "w")) == 0) {
		i_report(i_EFWRITE, fname);
		return false;
	}
	i_report(i_TLST, fname);
	return true;
}
