/*******************************************************************************
* Implementation of a MS-Windows GUI to visualise/debug the parsing proces of a
* parser generated with SLADE
*
* JHG sep97 SLADE Copyright (C) 1995-1997 University of Twente
*******************************************************************************/
#include "cgl.h"

#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
	extern char**	_argv;
	extern int		_argc;
#endif

void					i_next_token(void);

#define FILE_MENU		0
#define BREAK_MENU		1
#define STEP_ITEM		2
#define RUN_ITEM		3
#define RESET_ITEM		4
#define MAX_MENU_ROWS	20	/* max. nr. of menu items per menu collumn */
#define FIRSTNONTERM	0x400
#define FIRSTTERM		0x800

#define		RHS_SIZE	8092

extern i_NontermInfo	i_NtI[];
void					i_reset_cgl_files(void);
void					i_reset_vimcode(void);
void					i_reset_idscope(void);

static HWND				MainWnd, RunWnd;
static HINSTANCE		MainInst;
static int				StepWise, NontNr, PrevNontNr= -1;
static char				*Argv[2];
static jmp_buf			ResetBuf;
static int				ColNr, LineNr;	/* position in IdInput text field */
static int				NrInputLines;
static char				GrammarRule[RHS_SIZE];
static char				HelpPath[256];
static string			ReadStringBuf;
static BOOL				FromBegin;
static int				*TermBreak= 0;

#define MAXLINELEN		120

static void
ResetParser(int ac, char **av)
{
	LineNr= 0;
	ColNr= 0;

	SendDlgItemMessage(RunWnd, IdOutput, LB_RESETCONTENT, 0, 0);
	SendDlgItemMessage(RunWnd, IdNonterms, LB_RESETCONTENT, 0, 0);
	SendDlgItemMessage(RunWnd, IdAttrs, LB_RESETCONTENT, 0, 0);
	SendDlgItemMessage(RunWnd, IdInstr, LB_RESETCONTENT, 0, 0);
	SendDlgItemMessage(RunWnd, IdLenTab, LB_RESETCONTENT, 0, 0);
	SendDlgItemMessage(RunWnd, IdStringTab, LB_RESETCONTENT, 0, 0);
	EnableWindow(GetDlgItem(RunWnd, IdBreak), TRUE);
	EnableWindow(GetDlgItem(RunWnd, IdStep), TRUE);
	EnableWindow(GetDlgItem(RunWnd, IdRun), TRUE);
	EnableWindow(GetDlgItem(RunWnd, IdReset), FALSE);
	SetDlgItemText(RunWnd, IdGrammar, "");
	SetDlgItemText(RunWnd, IdInputTitle, "No tokens read");
	PrevNontNr= -1;
	StepWise= 0;

	i_reset_cgl_files();
	i_reset_vimcode();
	i_reset_idscope();
	FromBegin= TRUE;

	if (ac == 2) {
		i_init(ac, av);
		i_next_token();
	}
	SetFocus(GetDlgItem(RunWnd, IdStep));
}

static BOOL
IsCommand(MSG msg, int cmdId, char shortcut)
{
	if (msg.message == WM_LBUTTONUP && msg.hwnd == GetDlgItem(RunWnd, cmdId))
		return TRUE;

	if (msg.message == WM_CHAR) {
		if (shortcut == (char)msg.wParam)
			return TRUE;
		if (msg.wParam == ' ' && msg.hwnd == GetDlgItem(RunWnd, cmdId))
			return TRUE;
	}
	return FALSE;
}

static void
UserInput(void)
{
	MSG		msg;

	while (GetMessage(&msg, 0, 0, 0)) {
		if (IsDialogMessage(RunWnd, &msg)) {
			if (IsCommand(msg, IdClose, 'c'))
				break;

			if (IsCommand(msg, IdReset, 'e'))
				longjmp(ResetBuf, 1);

			if (IsCommand(msg, IdStep, 's')) {
				StepWise= 1;
				return;
			}

			if (IsCommand(msg, IdRun, 'r')) {
				StepWise= 0;
				return;
			}
		} else {
			TranslateMessage(&msg);		// Translates virtual key codes
			DispatchMessage(&msg); // Dispatches message to window
		}
	}
	exit(0);						// got an IdClose, so: Bye bye
}

void
i_add_list_item(int id, char *fmt, ...)
{
	va_list	args;
	string	buf;
	int		i;

	va_start(args, fmt);
	vsprintf(buf, fmt, args);
	va_end(args);

	if (id < 0) {
		id= -id;
		i= (int)SendDlgItemMessage(RunWnd, id, LB_ADDSTRING, 0, (LPARAM) buf);
		SendDlgItemMessage(RunWnd, id, LB_SETCURSEL, i, 0);
		UserInput();
	} else {
		i= (int)SendDlgItemMessage(RunWnd, id, LB_ADDSTRING, 0, (LPARAM) buf);
		SendDlgItemMessage(RunWnd, id, LB_SETCURSEL, i, 0);
	}
}

static void
DeleteListItem(int id)
{
	int	i= (int)SendDlgItemMessage(RunWnd, id, LB_GETCOUNT, 0, 0);
	SendDlgItemMessage(RunWnd, id, LB_DELETESTRING, i - 1, 0);
	SendDlgItemMessage(RunWnd, id, LB_SETCURSEL, i - 2, 0);
}

static void
ShowGrammarRule(int nr)
{
	FILE		*fp;
	char		*sep, buf[MAXLINELEN+1];
	int			len, found= 0;

	if (nr == PrevNontNr)
		return;
	PrevNontNr= nr;

	if ((sep= strrchr(Argv[0], '\\')) != 0)
		strcpy(buf, sep+1);
	else	
		strcpy(buf, Argv[0]);
	
	if ((sep= strrchr(buf, '.')) != 0)
		*(sep-1)= 0;			/* remove the trailing d.exe */
	else
		buf[strlen(buf)-2]= 0;	/* just remove the trailing d */

	strcat(buf, "_p.lst");
	if ((fp= fopen(buf, "rt")) == 0) {
		sprintf(GrammarRule, "Cannot open '%s'", buf);
		SetDlgItemText(RunWnd, IdGrammar, GrammarRule);
		return;
	}

	GrammarRule[0]= 0;
	len= strlen(i_NtI[nr].name);
	while (fgets(buf, MAXLINELEN, fp) != 0) {
		if (!found) {
			if (buf[0] == '\t' || buf[0] == ' ')
				continue;
			found= !strncmp(buf, i_NtI[nr].name, len)
			&& (buf[len] == ':' || buf[len] == ' ' || buf[len] == '\n');
		}
		if (found) {
			len= strlen(buf) + 1;
			if (strlen(GrammarRule) + len > RHS_SIZE)
				break;
			buf[len-2]= '\r';
			buf[len-1]= '\n';
			buf[len]= '\0';
			strcat(GrammarRule, buf);

			if (buf[0] == '.')
				break;
		}
	}
	SetDlgItemText(RunWnd, IdGrammar, GrammarRule);
	fclose(fp);
}

void
i_pass_term(int iterm, ...)
{
	va_list	args;
	char	*attr, *value;
	int		nattrs= 0;

	va_start(args, iterm);
	while ((attr= va_arg(args, char *)) != 0) {
		value= va_arg(args, char *);
		i_add_list_item(IdAttrs, "%s of %s: %s", attr, i_token_names[iterm]
		, value);
		nattrs++;
	}
	va_end(args);

	if (TermBreak[iterm]) {
		ShowGrammarRule(NontNr);
		UserInput();
	}

	while (nattrs--)
		DeleteListItem(IdAttrs);
}

void
i_enter_nonterm(int nr, ...)
{
	va_list	args;
	char	*attr, *value;

	va_start(args, nr);
	while ((attr= va_arg(args, char *)) != 0) {
		value= va_arg(args, char *);
		i_add_list_item(IdAttrs, "%s of %s: %s", attr, i_NtI[nr].name, value);
	}
	va_end(args);
	
	i_add_list_item(IdNonterms, i_NtI[nr].name);
	NontNr= nr;

	if (i_NtI[nr].brk) {
		ShowGrammarRule(nr);
		UserInput();
	}
}

static int
GetNonterminalNr(int isel)
{
	string	name;
	int		i;

	SendDlgItemMessage(RunWnd, IdNonterms, LB_GETTEXT, (WPARAM)isel
	, (LPARAM)name);
	for (i= 0; i_NtI[i].name != 0; i++)
		if (!strcmp(name, i_NtI[i].name))
			return i;
	return 0;
}

static int
GetLastNonterminalNr(void)
{
	int		n= (int)SendDlgItemMessage(RunWnd, IdNonterms, LB_GETCOUNT, 0, 0);

	return GetNonterminalNr(n-1);
}

void
i_exit_nonterm(int nr, ...)
{
	va_list	args;
	char	*attr, *value;
	int		nattrs= i_NtI[nr].nattrs;

	va_start(args, nr);
	while ((attr= va_arg(args, char *)) != 0) {
		value= va_arg(args, char *);
		i_add_list_item(IdAttrs,"%s of %s: %s", attr, i_NtI[nr].name, value);
	}
	va_end(args);
	
	if (i_NtI[nr].brk) {
		ShowGrammarRule(nr);
		UserInput();
	}

	DeleteListItem(IdNonterms);
	while (nattrs--)
		DeleteListItem(IdAttrs);
	NontNr= GetLastNonterminalNr();
}

void
i_select_token(char *pattern, char *token)
{
	char		*match, linebuf[MAXLINELEN+1];
	int			len, linelen, offset;

	if (StepWise && NontNr != PrevNontNr) {
		ShowGrammarRule(NontNr);
		UserInput();
	}

	linebuf[0]= MAXLINELEN;
	linelen= (int)SendDlgItemMessage(RunWnd, IdInput, EM_GETLINE, LineNr
	, (LPARAM)linebuf);

	if (linelen >= MAXLINELEN || pattern == 0 || !*pattern)
		return;

	linebuf[linelen]= 0;

	if ((match= strstr(linebuf + ColNr, pattern)) == 0) {
		do {
			if (++LineNr == NrInputLines) {
				LineNr--;
				return;
			}
			linebuf[0]= MAXLINELEN;
			linelen= (int)SendDlgItemMessage(RunWnd, IdInput, EM_GETLINE, LineNr
			, (LPARAM)linebuf);
			linebuf[linelen]= 0;
		} while ((match= strstr(linebuf, pattern)) == 0);
		ColNr= 0;
	}

	len= strlen(pattern);
	ColNr= (int)(match - linebuf) + len;

	offset= (int)SendDlgItemMessage(RunWnd, IdInput, EM_LINEINDEX, LineNr, 0)
	+ ColNr;
#ifdef WIN32
	SendDlgItemMessage(RunWnd, IdInput, EM_SETSEL, offset-len, offset);
	SendDlgItemMessage(RunWnd, IdInput, EM_SCROLLCARET, 0, 0);
#else /* WIN16 */
	{
		LPARAM	lParam= MAKELPARAM(offset-len, offset);
		SendDlgItemMessage(RunWnd, IdInput, EM_SETSEL, 0, lParam);
	}
#endif
	sprintf(linebuf, "Token '%s' read", token);
	SetDlgItemText(RunWnd, IdInputTitle, linebuf);

	if (StepWise)
		UserInput();
}

static BOOL
ReadInputFile(char *path)
{
	PSTR	filebufp;
	HLOCAL	filebuf;
	string	title;


#ifdef WIN32
	HANDLE			fh;
	DWORD			size, nread;

	if ((fh= CreateFile(path, GENERIC_READ, FILE_SHARE_READ, 0
	, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0)) == INVALID_HANDLE_VALUE)
		return FALSE;

	if ((size= GetFileSize(fh, 0)) == 0xFFFFFFFF) {
		CloseHandle(fh);
		return FALSE;
	}

	filebuf= LocalAlloc(LMEM_MOVEABLE|LMEM_ZEROINIT, size+1);
	if (filebuf == 0) {
		CloseHandle(fh);
		return FALSE;
	}
	
	filebufp= LocalLock(filebuf);
	if (ReadFile(fh, filebufp, size, &nread, 0) == FALSE
	|| nread != size) {
		LocalUnlock(filebuf);
		CloseHandle(fh);
		return FALSE;
	}
	CloseHandle(fh);
#else
	struct stat		statbuf;
	int				fd;
	OFSTRUCT		of;
	UINT			size;

	of.cBytes= sizeof(OFSTRUCT);
	if ((fd= OpenFile(path, &of, OF_READ)) < 0)
		return FALSE;

	if (fstat(fd, &statbuf) < 0) {
		close(fd);
		return FALSE;
	}
	size= (UINT)statbuf.st_size;

	filebuf= LocalAlloc(LMEM_MOVEABLE|LMEM_ZEROINIT, size+1);
	if (filebuf == 0) {
		close(fd);
		return FALSE;
	}
	
	filebufp= LocalLock(filebuf);
	if (read(fd, (char *)filebufp, (int)statbuf.st_size)
	!= (int)statbuf.st_size) {
		LocalUnlock(filebuf);
		close(fd);
		return FALSE;
	}
	close(fd);
#endif /* WIN32 */

	LocalUnlock(filebuf);
	SendDlgItemMessage(RunWnd, IdInput, EM_SETHANDLE, (WPARAM)filebuf, 0);
	NrInputLines= (int)SendDlgItemMessage(RunWnd, IdInput, EM_GETLINECOUNT,0,0);

	if (Argv[1] != 0)
		free(Argv[1]);
	Argv[1]= strdup(path);

	ResetParser(2, Argv);
	sprintf(title, "%s - %s", Argv[0], path);
	SetWindowText(MainWnd, title);
	return TRUE;
}

static void
BreakMenu(HMENU menu)
{
	int		i, maxmenu, nitems;
	HMENU	popup;

	popup= CreatePopupMenu();
	for (nitems= 0; i_token_names[nitems] != 0; nitems++)
		;

	maxmenu= nitems--;			 /* determine nr. of collumns in menu */
	for (i= 1; maxmenu > MAX_MENU_ROWS; i++)
		maxmenu= (nitems / i) + 1;

	for (i= 0; i < nitems; i++) {
		if (i && i % maxmenu == 0)
			AppendMenu(popup, MF_STRING|MF_MENUBARBREAK, FIRSTTERM+i
			, i_token_names[i+1]);
		else
			AppendMenu(popup, MF_STRING, FIRSTTERM+i, i_token_names[i+1]);

		if (TermBreak[i+1])
			CheckMenuItem(popup, FIRSTTERM+i, MF_BYCOMMAND | MF_CHECKED);
		else
			CheckMenuItem(popup, FIRSTTERM+i, MF_BYCOMMAND | MF_UNCHECKED);
	}
	AppendMenu(menu, MF_STRING|MF_POPUP, (UINT)popup, "Terminal");

	popup= CreatePopupMenu();
	for (nitems= 0; i_NtI[nitems].name != 0; nitems++)
		;

	maxmenu= nitems;			 /* determine nr. of collumns in menu */
	for (i= 1; maxmenu > MAX_MENU_ROWS; i++)
		maxmenu= (nitems / i) + 1;

	for (i= 0; i < nitems; i++) {
		if (i && i % maxmenu == 0)
			AppendMenu(popup, MF_STRING|MF_MENUBARBREAK, FIRSTNONTERM+i
			, i_NtI[i].name);
		else
			AppendMenu(popup, MF_STRING, FIRSTNONTERM+i, i_NtI[i].name);

		if (i_NtI[i].brk)
			CheckMenuItem(popup, FIRSTNONTERM+i, MF_BYCOMMAND | MF_CHECKED);
		else
			CheckMenuItem(popup, FIRSTNONTERM+i, MF_BYCOMMAND | MF_UNCHECKED);
	}
	AppendMenu(menu, MF_STRING|MF_POPUP, (UINT)popup, "Nonterminal");

}

BOOL
ReadStringDlgProc(HWND hDlg, WORD wMsg, WORD wParam, DWORD lParam)
{

	if (wMsg == WM_INITDIALOG) {
		SetDlgItemText(hDlg, IdLabel, (LPCSTR)lParam);
		SetWindowText(hDlg, (LPCSTR)lParam);
		return TRUE;
	}

	if (wMsg == WM_COMMAND) {
#ifdef WIN32
		short	cmdId= LOWORD(wParam);
#else /* WIN16 */
		int		cmdId= wParam;
#endif
		switch(cmdId) {
		case IDCANCEL:
			ReadStringBuf[0]= 0;
			EndDialog(hDlg, FALSE);
			return TRUE;

		case IDOK:
			GetDlgItemText(hDlg, IdText, ReadStringBuf, i_STRINGLEN);
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
	}
	return FALSE;
}

char *
i_read_line_dialog(char *s, int len, char *prompt)
{
#ifdef WIN32
	DialogBoxParam(MainInst, "ReadString", RunWnd, (DLGPROC)ReadStringDlgProc
	, (LPARAM)prompt);
#else /* WIN16 */
	FARPROC	dlgprc= (FARPROC)MakeProcInstance((FARPROC)ReadStringDlgProc
	, MainInst);
	DialogBoxParam(MainInst, "ReadString", RunWnd, (DLGPROC)dlgprc, (LPARAM)prompt);
	FreeProcInstance((FARPROC)dlgprc);
#endif
	strncpy(s, ReadStringBuf, len-1);
	s[len-1]= 0;
	return s;
}

BOOL CALLBACK
RunDlgProc(HWND hDlg, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	HLOCAL	filebuf;
	HMENU	popup;
	RECT	rc;

	if (wMsg == WM_INITDIALOG) {
		char	*sladehome;

		SendDlgItemMessage(hDlg, IdOutput, LB_SETHORIZONTALEXTENT, 1500, 0L);
		SendDlgItemMessage(hDlg, IdAttrs, LB_SETHORIZONTALEXTENT, 1500, 0L);
		SendDlgItemMessage(hDlg, IdStringTab, LB_SETHORIZONTALEXTENT, 1500, 0L);
		if ((sladehome= getenv("SLADE")) != 0)
			sprintf(HelpPath, "%s/lib/slade.hlp", sladehome);
		else
			HelpPath[0]= 0;

		if (TermBreak == 0) {
			int		nitems;
			for (nitems= 0; i_token_names[nitems] != 0; nitems++)
				;
			TermBreak= (int *)calloc(nitems, sizeof(int));
		}

		return TRUE;
	}
	
	if (wMsg == WM_COMMAND) {
#ifdef WIN32
		short	cmdId= LOWORD(wParam);
		short	cmdNot= HIWORD(wParam);
#else /* WIN16 */
		int		cmdId= wParam;
		int		cmdNot= HIWORD(lParam);
#endif
		switch(cmdId) {
		case IdClose:
			filebuf= (HLOCAL)SendDlgItemMessage(hDlg, IdInput,EM_GETHANDLE,0,0);
			if (LocalSize(filebuf) > 0)
				LocalFree(filebuf);
			ResetParser(0, 0);
			DestroyWindow(MainWnd);
			return TRUE;

		case IdBreak:
			popup= CreatePopupMenu();
			BreakMenu(popup);
			GetWindowRect(GetDlgItem(hDlg, cmdId), &rc);
			TrackPopupMenu(popup, 0, rc.left, rc.bottom, 0, hDlg, 0);
			DestroyMenu(popup);
			return TRUE;

		case IdRun:
		case IdStep:
			StepWise= (cmdId == IdStep);
			if (FromBegin) {
				FromBegin= FALSE;
				EnableWindow(GetDlgItem(hDlg, IdReset), TRUE);
				p_ROOT();
				i_finish();
				EnableWindow(GetDlgItem(hDlg, IdStep), FALSE);
				EnableWindow(GetDlgItem(hDlg, IdRun), FALSE);
				SetDlgItemText(hDlg, IdGrammar, "");
				SetFocus(GetDlgItem(hDlg, IdReset));
			}
			return TRUE;

		case IdReset:
			ResetParser(2, Argv);
			return TRUE;

		case IdNonterms:
			if (cmdNot == LBN_SELCHANGE) {
				int	i= (int)SendDlgItemMessage(hDlg, cmdId, LB_GETCURSEL, 0, 0);
				ShowGrammarRule(GetNonterminalNr(i));
			}
			return TRUE;

		case IdHelp:
			if (HelpPath[0])
				WinHelp(hDlg, HelpPath, HELP_CONTEXT, HelpDebug);
			return TRUE;

		default:
			if (cmdId >= FIRSTTERM) {
				int		item= cmdId - FIRSTTERM;

				TermBreak[item+1]= !TermBreak[item+1];
				return TRUE;
			}
			if (cmdId >= FIRSTNONTERM) {
				int		item= cmdId - FIRSTNONTERM;

				i_NtI[item].brk= !i_NtI[item].brk;
				return TRUE;
			}
		}
		return FALSE;
	}
	return FALSE;					// Passes it on if unproccessed
}
 
LRESULT CALLBACK
WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HLOCAL	filebuf;

	switch (message) {
	case WM_CLOSE:
		ResetParser(0, 0);
		filebuf= (HLOCAL)SendDlgItemMessage(RunWnd, IdInput, EM_GETHANDLE, 0,0);
		if (LocalSize(filebuf) > 0)
			LocalFree(filebuf);
		DestroyWindow(hWnd);
		break;

	case WM_QUIT:
	case WM_DESTROY:				// message: window being destroyed
		PostQuitMessage(0);
		free(TermBreak);
		break;

	default:						// Passes it on if unproccessed
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

static BOOL
InitApplication(HINSTANCE hInst)
{
	WNDCLASS wc;

	wc.style= CS_HREDRAW | CS_VREDRAW;	// Class style(s)
	wc.lpfnWndProc= (WNDPROC) WndProc;
	wc.cbClsExtra= 0;					// No per-class extra data
	wc.cbWndExtra= 0;					// No per-window extra data
	wc.hInstance= hInst;				// Application that owns the class
	wc.hIcon= LoadIcon(0, IDI_APPLICATION);
	wc.hCursor= LoadCursor(0, IDC_ARROW);
	wc.hbrBackground= GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName = 0;		// Name of menu resource in .RC file.
	wc.lpszClassName= "MainWnd";		// Name used in call to CreateWindow

	return RegisterClass(&wc);
}

static BOOL
InitInstance(HINSTANCE hInst, int nCmdShow)
{
	RECT		rc;
	POINT		lr;

	MainInst= hInst;
	MainWnd= CreateWindow(
		"MainWnd",						// See RegisterClass() call
		Argv[0],						// Text for window title bar
		WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX,	// Window style
		CW_USEDEFAULT,					// horizontal pos
		CW_USEDEFAULT,					// vertical pos
		CW_USEDEFAULT,					// width
		CW_USEDEFAULT,					// height
		0,								// Overlapped, so no parent
		0,								// Use the window class menu
		MainInst,						// This instance is owner
		0
	);

	if (!MainWnd)
		return FALSE;

#ifdef WIN32
	RunWnd= CreateDialog(MainInst, "RunDialog", MainWnd, (DLGPROC)RunDlgProc);
#else /* WIN16 */
	{
	DLGPROC	dlgprc= (DLGPROC)MakeProcInstance((FARPROC)RunDlgProc, MainInst);
	RunWnd= CreateDialog(MainInst, "RUNDIALOG", MainWnd, (DLGPROC)dlgprc);
	}
#endif
	if (!RunWnd)
		return FALSE;

	GetWindowRect(RunWnd, &rc);
	lr.x= rc.right;
	lr.y= rc.bottom;
	ScreenToClient(MainWnd, &lr);
	SetWindowPos(MainWnd, 0,0,0, lr.x, lr.y+24, SWP_NOZORDER|SWP_NOMOVE);

	ShowWindow(MainWnd, nCmdShow);
	ShowWindow(RunWnd, SW_SHOW);
	return TRUE;
}

int WINAPI
WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	MSG		msg;
	char	*sep;
	

	if (!hPrevInst && !InitApplication(hInst))
		return FALSE;

	if (!InitInstance(hInst, nCmdShow))
		return FALSE;

#ifdef WIN32
	Argv[0]= GetCommandLine();
	if ((sep= strchr(Argv[0], ' ')) != 0 && *(sep+1)) {
		*sep= 0;
		if (!ReadInputFile(sep+1)) {
			sprintf(ReadStringBuf, "Cannot open %s", sep+1);
#else /* WIN16 */
	Argv[0]= _argv[0];
	if (_argc == 2) {
		sep= _argv[1];
		if (!ReadInputFile(sep)) {
			sprintf(ReadStringBuf, "Cannot open %s", sep);
#endif
			MessageBox(0, ReadStringBuf, 0, MB_OK|MB_APPLMODAL);
			return FALSE;
		}

	} else {
		MessageBox(0, "No input file specified", 0, MB_OK|MB_APPLMODAL);
		return FALSE;
	}

	setjmp(ResetBuf);
	while (GetMessage(&msg, 0, 0, 0)) {
		if (IsDialogMessage(RunWnd, &msg))
			continue;
		TranslateMessage(&msg);				// Translates virtual key codes
		DispatchMessage(&msg);				// Dispatches message to window
	}
	return msg.wParam;				// Returns the value from PostQuitMessage
}
